from pathlib import PurePath, PurePosixPath
from typing import Any

from pydantic import BaseModel, SecretStr


def default_serializer(value: Any) -> Any:
    """
    Args:
        value: A value to serialize
    Returns:
        A serialized value
    Raises:
        TypeError: if value is not supported
    """
    if isinstance(value, BaseModel):
        return value.dict()
    if isinstance(value, SecretStr):
        return value.get_secret_value()
    if isinstance(value, (PurePath, PurePosixPath)):
        return str(value)
    raise TypeError("unsupported type")



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Type, Union

from .asgi_types import ASGIApp
from .callable_types import ExceptionHandler

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo  # noqa: TC004
    from starlette.middleware import Middleware as StarletteMiddleware  # noqa: TC004
    from starlette.middleware.base import BaseHTTPMiddleware  # noqa: TC004

    from starlite.datastructures.cookie import Cookie  # noqa: TC004
    from starlite.datastructures.provide import Provide  # noqa: TC004
    from starlite.datastructures.response_header import ResponseHeader  # noqa: TC004
    from starlite.middleware.base import (  # noqa: TC004
        DefineMiddleware,
        MiddlewareProtocol,
    )

else:
    BaseHTTPMiddleware = Any
    Cookie = Any
    DefineMiddleware = Any
    FieldInfo = Any
    MiddlewareProtocol = Any
    Provide = Any
    ResponseHeader = Any
    StarletteMiddleware = Any


Dependencies = Dict[str, Provide]
ExceptionHandlersMap = Dict[Union[int, Type[Exception]], ExceptionHandler]

Middleware = Union[
    Callable[..., ASGIApp], DefineMiddleware, StarletteMiddleware, Type[BaseHTTPMiddleware], Type[MiddlewareProtocol]
]
ParametersMap = Dict[str, FieldInfo]
ResponseCookies = List[Cookie]
ResponseHeadersMap = Dict[str, ResponseHeader]



import io
from enum import Enum
from typing import TYPE_CHECKING, Optional, cast

from starlette.datastructures import Headers, MutableHeaders
from starlette.middleware.gzip import GZipResponder
from typing_extensions import Literal

from starlite.enums import ScopeType
from starlite.exceptions import MissingDependencyException
from starlite.middleware.base import MiddlewareProtocol

if TYPE_CHECKING:
    from starlite.types import ASGIApp, Message, Receive, Scope, Send
    from starlite.types.asgi_types import HTTPResponseStartEvent, HTTPScope

try:
    import brotli
except ImportError as e:
    raise MissingDependencyException("brotli is not installed") from e

BrotliMode = Literal["text", "generic", "font"]


class CompressionEncoding(str, Enum):
    """An Enum for supported compression encodings."""

    GZIP = "gzip"
    BROTLI = "br"


class BrotliMiddleware(MiddlewareProtocol):
    def __init__(
        self,
        app: "ASGIApp",
        minimum_size: int = 400,
        brotli_quality: int = 4,
        brotli_mode: BrotliMode = "text",
        brotli_lgwin: int = 22,
        brotli_lgblock: int = 0,
        brotli_gzip_fallback: bool = True,
    ) -> None:
        """Brotli middleware for Starlite.

        Compresses responses using Brotli and optional fallback to Gzip.

        Args:
            app: The 'next' ASGI app to call.
            minimum_size: Minimum size for the response body to affect compression.
            brotli_quality: Controls the compression-speed vs compression-density tradeoffs.
                The higher the quality, the slower the compression. The range of this value is 0 to 11.
            brotli_mode: The encoder mode.
            brotli_lgwin: The base-2 logarithm of the sliding window size. The range of this value is 10 to 24.
            brotli_lgblock: The base-2 logarithm of the maximum input block size. The range of this value is 16 to 24.
                If set to 0, the value will be set based on quality.
            brotli_gzip_fallback: Allow falling back to GZIP.
        """
        self.app = app
        self.quality = brotli_quality
        self.mode = self._brotli_mode_to_int(brotli_mode)
        self.minimum_size = minimum_size
        self.lgwin = brotli_lgwin
        self.lgblock = brotli_lgblock
        self.gzip_fallback = brotli_gzip_fallback

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        if scope["type"] == ScopeType.HTTP:
            headers = Headers(scope=scope)
            if CompressionEncoding.BROTLI in headers.get("Accept-Encoding", ""):
                brotli_responder = BrotliResponder(
                    app=self.app,
                    minimum_size=self.minimum_size,
                    quality=self.quality,
                    mode=self.mode,
                    lgwin=self.lgwin,
                    lgblock=self.lgblock,
                )
                await brotli_responder(scope, receive, send)
                return
            if self.gzip_fallback and CompressionEncoding.GZIP in headers.get("Accept-Encoding", ""):
                gzip_responder = GZipResponder(self.app, self.minimum_size)  # type: ignore[arg-type]
                await gzip_responder(scope, receive, send)  # type: ignore[arg-type]
                return
        await self.app(scope, receive, send)

    @staticmethod
    def _brotli_mode_to_int(brotli_mode: BrotliMode) -> int:
        """Select the correct brotli mode.

        Returns:
            An int correlating with the constants in the brotli package
        """
        if brotli_mode == "text":
            return int(brotli.MODE_TEXT)
        if brotli_mode == "font":
            return int(brotli.MODE_FONT)
        return int(brotli.MODE_GENERIC)


class BrotliResponder:
    def __init__(
        self,
        app: "ASGIApp",
        minimum_size: int,
        quality: int,
        mode: int,
        lgwin: int,
        lgblock: int,
    ) -> None:
        """Brotli Responder.

        Formats a response with Brotli compression.

        Args:
            app: The 'next' ASGI app to call.
            minimum_size: Minimum size for the response body to affect compression.
            quality: Controls the compression-speed vs compression-density tradeoffs.
                The higher the quality, the slower the compression. The range of this value is 0 to 11.
            mode: The encoder mode.
            lgwin: The base-2 logarithm of the sliding window size. The range of this value is 10 to 24.
            lgblock: The base-2 logarithm of the maximum input block size. The range of this value is 16 to 24.
                If set to 0, the value will be set based on quality.
        """
        self.app = app
        self.minimum_size = minimum_size
        self.initial_message: Optional["HTTPResponseStartEvent"] = None
        self.started = False
        self.br_buffer = io.BytesIO()
        self.quality = quality
        self.mode = mode
        self.lgwin = lgwin
        self.lgblock = lgblock
        self.br_file = brotli.Compressor(quality=self.quality, mode=self.mode, lgwin=self.lgwin, lgblock=self.lgblock)

    async def __call__(self, scope: "HTTPScope", receive: "Receive", send: "Send") -> None:
        await self.app(scope, receive, self.create_send_wrapper(send))

    def create_send_wrapper(self, send: "Send") -> "Send":
        """Wraps 'send' to handle brotli compression.

        Args:
            send: The ASGI send function.

        Returns:
            An ASGI send function.
        """

        async def send_wrapper(message: "Message") -> None:
            """Handles and compresses the HTTP Message with brotli.

            Args:
                message (Message): An ASGI Message.
            """

            if message["type"] == "http.response.start":
                # Don't send the initial message until we've determined how to
                # modify the outgoing headers correctly.
                self.initial_message = message
                return

            initial_message = cast("HTTPResponseStartEvent", self.initial_message)

            if message["type"] == "http.response.body" and not self.started:
                self.started = True
                body = message.get("body", b"")
                more_body = message.get("more_body", False)
                if len(body) < self.minimum_size and not more_body:
                    # Don't apply Brotli to small outgoing responses.
                    await send(initial_message)
                    await send(message)
                elif not more_body:
                    # Standard Brotli response.
                    body = self.br_file.process(body) + self.br_file.finish()
                    headers = MutableHeaders(raw=initial_message["headers"])
                    headers["Content-Encoding"] = CompressionEncoding.BROTLI
                    headers["Content-Length"] = str(len(body))
                    headers.add_vary_header("Accept-Encoding")
                    message["body"] = body

                    await send(initial_message)
                    await send(message)
                else:
                    # Initial body in streaming Brotli response.
                    headers = MutableHeaders(raw=initial_message["headers"])
                    headers["Content-Encoding"] = CompressionEncoding.BROTLI
                    headers.add_vary_header("Accept-Encoding")
                    del headers["Content-Length"]
                    self.br_buffer.write(self.br_file.process(body) + self.br_file.flush())
                    message["body"] = self.br_buffer.getvalue()
                    self.br_buffer.seek(0)
                    self.br_buffer.truncate()
                    await send(initial_message)
                    await send(message)
            elif message["type"] == "http.response.body":
                # Remaining body in streaming Brotli response.
                body = message.get("body", b"")
                more_body = message.get("more_body", False)
                self.br_buffer.write(self.br_file.process(body) + self.br_file.flush())
                if not more_body:
                    self.br_buffer.write(self.br_file.finish())

                message["body"] = self.br_buffer.getvalue()

                self.br_buffer.seek(0)
                self.br_buffer.truncate()
                if not more_body:
                    self.br_buffer.close()
                await send(message)

        return send_wrapper



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union, cast

from starlite.app import DEFAULT_CACHE_CONFIG, Starlite
from starlite.exceptions import MissingDependencyException
from starlite.middleware.session import SessionMiddleware

if TYPE_CHECKING:
    from typing_extensions import Literal

    from starlite import Request, WebSocket
    from starlite.config import (
        BaseLoggingConfig,
        CacheConfig,
        CompressionConfig,
        CORSConfig,
        CSRFConfig,
        OpenAPIConfig,
        StaticFilesConfig,
        TemplateConfig,
    )
    from starlite.middleware.session import SessionCookieConfig
    from starlite.plugins.base import PluginProtocol
    from starlite.types import (
        AfterExceptionHookHandler,
        AfterRequestHookHandler,
        AfterResponseHookHandler,
        BeforeMessageSendHookHandler,
        BeforeRequestHookHandler,
        ControllerRouterHandler,
        Dependencies,
        ExceptionHandlersMap,
        Guard,
        LifeSpanHandler,
        LifeSpanHookHandler,
        Middleware,
        ParametersMap,
        ResponseType,
        SingleOrList,
    )

try:
    from starlette.testclient import TestClient as StarletteTestClient
except ImportError as e:
    raise MissingDependencyException(
        "To use starlite.testing, install starlite with 'testing' extra, e.g. `pip install starlite[testing]`"
    ) from e


class TestClient(StarletteTestClient):
    app: Starlite  # type: ignore[assignment]
    """
        Starlite application instance under test.
    """

    def __init__(
        self,
        app: Starlite,
        base_url: str = "http://testserver",
        raise_server_exceptions: bool = True,
        root_path: str = "",
        backend: "Literal['asyncio', 'trio' ]" = "asyncio",
        backend_options: Optional[Dict[str, Any]] = None,
        session_config: Optional["SessionCookieConfig"] = None,
    ) -> None:
        """A client implementation providing a context manager for testing
        applications.

        Args:
            app: The instance of [Starlite][starlite.app.Starlite] under test.
            base_url: URL scheme and domain for test request paths, e.g. 'http://testserver'.
            raise_server_exceptions: Flag for underlying Starlette test client to raise server exceptions instead of
                wrapping them in an HTTP response.
            root_path: Path prefix for requests.
            backend: The async backend to use, options are "asyncio" or "trio".
            backend_options: 'anyio' options.
            session_config: Configuration for Session Middleware class to create raw session cookies for request to the
                route handlers.
        """
        self.session = SessionMiddleware(app=app, config=session_config) if session_config else None
        super().__init__(
            app=app,  # type: ignore[arg-type]
            base_url=base_url,
            raise_server_exceptions=raise_server_exceptions,
            root_path=root_path,
            backend=backend,
            backend_options=backend_options,
        )

    def __enter__(self) -> "TestClient":
        """Starlette's `TestClient.__enter__()` return value is strongly typed
        to return their own `TestClient`, i.e., not-generic to support
        subclassing.

        We override here to provide a nicer typing experience for our user

        Returns:
            TestClient
        """
        return super().__enter__()  # pyright: ignore

    def create_session_cookies(self, session_data: Dict[str, Any]) -> Dict[str, str]:
        """Creates raw session cookies that are loaded into the session by the
        Session Middleware. It creates cookies the same way as if they are
        coming from the browser. Your tests must set up session middleware to
        load raw session cookies into the session.

        Args:
            session_data: Dictionary to create raw session cookies from.

        Returns:
            A dictionary with cookie name as key and cookie value as value.

        Examples:

            ```python
            import pytest
            from starlite.testing import TestClient

            from my_app.main import app, session_cookie_config_instance


            class TestClass:
                @pytest.fixture()
                def test_client(self) -> TestClient:
                    with TestClient(
                        app=app, session_config=session_cookie_config_instance
                    ) as client:
                        yield client

                def test_something(self, test_client: TestClient) -> None:
                    cookies = test_client.create_session_cookies(session_data={"user": "test_user"})
                    # Set raw session cookies to the "cookies" attribute of test_client instance.
                    test_client.cookies = cookies
                    test_client.get(url="/my_route")
            ```
        """
        if self.session is None:
            return {}
        encoded_data = self.session.dump_data(data=session_data)
        return {f"{self.session.config.key}-{i}": chunk.decode("utf-8") for i, chunk in enumerate(encoded_data)}

    def get_session_from_cookies(self) -> Dict[str, Any]:
        """Raw session cookies are a serialized image of session which are
        created by session middleware and sent with the response. To assert
        data in session, this method deserializes the raw session cookies and
        creates session from them.

        Returns:
            A dictionary containing session data.

        Examples:

            ```python
            def test_something(self, test_client: TestClient) -> None:
                test_client.get(url="/my_route")
                session = test_client.get_session_from_cookies()
                assert "user" in session
            ```
        """
        if self.session is None:
            return {}
        raw_data = [self.cookies[key].encode("utf-8") for key in self.cookies if self.session.config.key in key]
        return self.session.load_data(data=raw_data)


def create_test_client(
    route_handlers: Union["ControllerRouterHandler", List["ControllerRouterHandler"]],
    *,
    after_exception: Optional["SingleOrList[AfterExceptionHookHandler]"] = None,
    after_request: Optional["AfterRequestHookHandler"] = None,
    after_response: Optional["AfterResponseHookHandler"] = None,
    after_shutdown: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
    after_startup: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
    allowed_hosts: Optional[List[str]] = None,
    backend: "Literal['asyncio', 'trio']" = "asyncio",
    backend_options: Optional[Dict[str, Any]] = None,
    base_url: str = "http://testserver",
    before_request: Optional["BeforeRequestHookHandler"] = None,
    before_send: Optional["SingleOrList[BeforeMessageSendHookHandler]"] = None,
    before_shutdown: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
    before_startup: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
    cache_config: "CacheConfig" = DEFAULT_CACHE_CONFIG,
    compression_config: Optional["CompressionConfig"] = None,
    cors_config: Optional["CORSConfig"] = None,
    csrf_config: Optional["CSRFConfig"] = None,
    dependencies: Optional["Dependencies"] = None,
    exception_handlers: Optional["ExceptionHandlersMap"] = None,
    guards: Optional[List["Guard"]] = None,
    logging_config: Optional["BaseLoggingConfig"] = None,
    middleware: Optional[List["Middleware"]] = None,
    on_shutdown: Optional[List["LifeSpanHandler"]] = None,
    on_startup: Optional[List["LifeSpanHandler"]] = None,
    openapi_config: Optional["OpenAPIConfig"] = None,
    parameters: Optional["ParametersMap"] = None,
    plugins: Optional[List["PluginProtocol"]] = None,
    raise_server_exceptions: bool = True,
    request_class: Optional[Type["Request"]] = None,
    response_class: Optional["ResponseType"] = None,
    root_path: str = "",
    session_config: Optional["SessionCookieConfig"] = None,
    static_files_config: Optional[Union["StaticFilesConfig", List["StaticFilesConfig"]]] = None,
    template_config: Optional["TemplateConfig"] = None,
    websocket_class: Optional[Type["WebSocket"]] = None,
) -> TestClient:
    """Creates a Starlite app instance and initializes it.

    [TestClient][starlite.testing.TestClient] with it.

    Notes:
        - This function should be called as a context manager to ensure async startup and shutdown are
            handled correctly.

    Examples:

        ```python
        from starlite import get, create_test_client


        @get("/some-path")
        def my_handler() -> dict[str, str]:
            return {"hello": "world"}


        def test_my_handler() -> None:
            with create_test_client(my_handler) as client:
                response == client.get("/some-path")
                assert response.json() == {"hello": "world"}
        ```

    Args:
        route_handlers: A single handler or a list of route handlers, which can include instances of
            [Router][starlite.router.Router], subclasses of [Controller][starlite.controller.Controller] or
            any function decorated by the route handler decorators.
        after_exception: An application level [exception event handler][starlite.types.AfterExceptionHookHandler].
            This hook is called after an exception occurs. In difference to exception handlers, it is not meant to
            return a response - only to process the exception (e.g. log it, send it to Sentry etc.).
        after_request: A sync or async function executed after the route handler function returned and the response
            object has been resolved. Receives the response object which may be either an instance of
            [Response][starlite.response.Response] or `starlette.Response`.
        after_response: A sync or async function called after the response has been awaited. It receives the
            [Request][starlite.connection.Request] object and should not return any values.
        after_shutdown: An application level [LifeSpan hook handler][starlite.types.LifeSpanHookHandler].
            This hook is called during the ASGI shutdown, after all callables in the 'on_shutdown'
            list have been called.
        after_startup: An application level [LifeSpan hook handler][starlite.types.LifeSpanHookHandler].
            This hook is called during the ASGI startup, after all callables in the 'on_startup'
            list have been called.
        allowed_hosts: A list of allowed hosts - enables the builtin allowed hosts middleware.
        backend: The async backend to use, options are "asyncio" or "trio".
        backend_options: 'anyio' options.
        base_url: URL scheme and domain for test request paths, e.g. 'http://testserver'.
        before_request: A sync or async function called immediately before calling the route handler.
            Receives the [Request][starlite.connection.Request] instance and any non-`None` return value is
            used for the response, bypassing the route handler.
        before_send: An application level [before send hook handler][starlite.types.BeforeMessageSendHookHandler] or
            list thereof. This hook is called when the ASGI send function is called.
        before_shutdown: An application level [LifeSpan hook handler][starlite.types.LifeSpanHookHandler]. This hook is
            called during the ASGI shutdown, before any callables in the 'on_shutdown' list have been called.
        before_startup: An application level [LifeSpan hook handler][starlite.types.LifeSpanHookHandler]. This hook is
            called during the ASGI startup, before any callables in the 'on_startup' list have been called.
        cache_config: Configures caching behavior of the application.
        compression_config: Configures compression behaviour of the application, this enabled a builtin or user
            defined Compression middleware.
        cors_config: If set this enables the builtin CORS middleware.
        csrf_config: If set this enables the builtin CSRF middleware.
        dependencies: A string keyed dictionary of dependency [Provider][starlite.datastructures.Provide] instances.
        exception_handlers: A dictionary that maps handler functions to status codes and/or exception types.
        guards: A list of [Guard][starlite.types.Guard] callables.
        logging_config: A subclass of [BaseLoggingConfig][starlite.config.logging.BaseLoggingConfig].
        middleware: A list of [Middleware][starlite.types.Middleware].
        on_shutdown: A list of [LifeSpanHandler][starlite.types.LifeSpanHandler] called during
            application shutdown.
        on_startup: A list of [LifeSpanHandler][starlite.types.LifeSpanHandler] called during
            application startup.
        openapi_config: Defaults to [DEFAULT_OPENAPI_CONFIG][starlite.app.DEFAULT_OPENAPI_CONFIG]
        parameters: A mapping of [Parameter][starlite.params.Parameter] definitions available to all
            application paths.
        plugins: List of plugins.
        request_class: An optional subclass of [Request][starlite.connection.request.Request] to use for
            http connections.
        raise_server_exceptions: Flag for underlying Starlette test client to raise server exceptions instead of
            wrapping them in an HTTP response.
        response_class: A custom subclass of [starlite.response.Response] to be used as the app's default response.
        root_path: Path prefix for requests.
        static_files_config: An instance or list of [StaticFilesConfig][starlite.config.StaticFilesConfig]
        session_config: Configuration for Session Middleware class to create raw session cookies for request to the
            route handlers.
        template_config: An instance of [TemplateConfig][starlite.config.TemplateConfig]
        websocket_class: An optional subclass of [WebSocket][starlite.connection.websocket.WebSocket] to use for
            websocket connections.

    Returns:
        An instance of [TestClient][starlite.testing.TestClient] with a created app instance.
    """
    return TestClient(
        app=Starlite(
            after_exception=after_exception,
            after_request=after_request,
            after_response=after_response,
            after_shutdown=after_shutdown,
            after_startup=after_startup,
            allowed_hosts=allowed_hosts,
            before_request=before_request,
            before_send=before_send,
            before_shutdown=before_shutdown,
            before_startup=before_startup,
            cache_config=cache_config,
            compression_config=compression_config,
            cors_config=cors_config,
            csrf_config=csrf_config,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            guards=guards,
            logging_config=logging_config,
            middleware=middleware,
            on_shutdown=on_shutdown,
            on_startup=on_startup,
            openapi_config=openapi_config,
            parameters=parameters,
            plugins=plugins,
            request_class=request_class,
            response_class=response_class,
            route_handlers=cast("Any", route_handlers if isinstance(route_handlers, list) else [route_handlers]),
            static_files_config=static_files_config,
            template_config=template_config,
            websocket_class=websocket_class,
        ),
        backend=backend,
        backend_options=backend_options,
        base_url=base_url,
        raise_server_exceptions=raise_server_exceptions,
        root_path=root_path,
        session_config=session_config,
    )



from .csrf import generate_csrf_hash, generate_csrf_token
from .dependency import is_dependency_field, should_skip_dependency_validation
from .exception import (
    ExceptionResponseContent,
    create_exception_response,
    get_exception_handler,
)
from .extractors import ConnectionDataExtractor, ResponseDataExtractor, obfuscate
from .model import convert_dataclass_to_model, create_parsed_model_field
from .path import join_paths, normalize_path
from .predicates import is_async_callable, is_class_and_subclass, is_optional_union
from .scope import get_serializer_from_scope
from .sequence import find_index, unique
from .serialization import default_serializer
from .sync import AsyncCallable, as_async_callable_list, async_partial

__all__ = (
    "AsyncCallable",
    "ConnectionDataExtractor",
    "ExceptionResponseContent",
    "ResponseDataExtractor",
    "as_async_callable_list",
    "async_partial",
    "convert_dataclass_to_model",
    "create_exception_response",
    "create_parsed_model_field",
    "default_serializer",
    "find_index",
    "get_exception_handler",
    "get_serializer_from_scope",
    "is_async_callable",
    "is_class_and_subclass",
    "is_dependency_field",
    "is_optional_union",
    "join_paths",
    "normalize_path",
    "obfuscate",
    "should_skip_dependency_validation",
    "unique",
    "generate_csrf_hash",
    "generate_csrf_token",
)



import pickle
from typing import Any, Callable, Optional

from pydantic import BaseModel

from starlite.exceptions import MissingDependencyException

try:
    from aiomcache import Client
except ImportError as e:
    raise MissingDependencyException(
        "To use starlite.memcached_cache_backend, install starlite with 'memcached' extra, e.g. `pip install starlite[memcached]`"
    ) from e


from starlite.cache.base import CacheBackendProtocol


class MemcachedCacheBackendConfig(BaseModel):
    """Memcached cache backend configuration."""

    host: str
    """memcached host"""
    port: Optional[int] = None
    """memcached port (optional, defaults to 11211)"""
    pool_size: Optional[int] = None
    """Maximum number of memcached open connections (optional, defaults to 2)"""
    pool_minsize: Optional[int] = None
    """memcached minimum pool size (optional, by default set to `pool_size`)"""
    serialize: Callable[[Any], bytes] = pickle.dumps
    """A callable to serialize data that goes into the cache from an object to bytes, defaults to `pickle.dumps`"""
    deserialize: Callable[[bytes], Any] = pickle.loads
    """A callable to deserialize data coming from the cache from bytes to an object, defaults to `pickle.loads`"""


class MemcachedCacheBackend(CacheBackendProtocol):
    def __init__(self, config: MemcachedCacheBackendConfig):
        """This class offers a cache backend based on memcached.

        Args:
            config: required configuration to connect to memcached.
        """
        self._config = config
        self._client: Client = None  # pyright: ignore

    @property
    def _memcached_client(self) -> Client:
        if not self._client:
            self._client = Client(**self._config.dict(exclude_unset=True))

        return self._client

    async def get(self, key: str) -> Any:  # pylint: disable=invalid-overridden-method
        """Retrieves a value from cache corresponding to the given key.

        Args:
            key: name of cached value.

        Returns:
            Cached value if existing else `None`.
        """

        value = await self._memcached_client.get(key=key.encode())  # pyright: ignore
        return self._config.deserialize(value)

    async def set(self, key: str, value: Any, expiration: int) -> None:  # pylint: disable=invalid-overridden-method
        """Set sa value in cache for a given key for a duration determined by
        expiration.

        Args:
            key: key to cache `value` under.
            value: the value to be cached.
            expiration: expiration of cached value in seconds.

        Notes:
            - expiration is in seconds.
            - return value is not used by Starlite internally.

        Returns:
            None
        """

        await self._memcached_client.set(key.encode(), self._config.serialize(value), exptime=expiration)

    async def delete(self, key: str) -> None:  # pylint: disable=invalid-overridden-method
        """Deletes a value from the cache and removes the given key.

        Args:
            key: key to be deleted from the cache.

        Notes:
            - return value is not used by Starlite internally.

        Returns:
            None
        """

        await self._memcached_client.delete(key=key.encode())



from typing import Any, Callable, List, TypeVar

from starlette.background import BackgroundTask as StarletteBackgroundTask
from starlette.background import BackgroundTasks as StarletteBackgroundTasks
from typing_extensions import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


class BackgroundTask(StarletteBackgroundTask):
    def __init__(self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> None:
        """A container for a 'background' task function. Background tasks are
        called once a Response finishes.

        Args:
            func: A sync or async function to call as the background task.
            *args: Args to pass to the func.
            **kwargs: Kwargs to pass to the func
        """
        super().__init__(func, *args, **kwargs)


class BackgroundTasks(StarletteBackgroundTasks):
    def __init__(self, tasks: List[BackgroundTask]) -> None:
        """A container for multiple 'background' task functions. Background
        tasks are called once a Response finishes.

        Args:
            tasks: A list of [BackgroundTask][starlite.datastructures.BackgroundTask] instances.
        """
        super().__init__(tasks=tasks)



from typing import TYPE_CHECKING

from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from starlite.connection import Request
from starlite.enums import ScopeType
from starlite.exceptions import WebSocketException
from starlite.utils import create_exception_response
from starlite.utils.exception import get_exception_handler

if TYPE_CHECKING:

    from starlette.responses import Response as StarletteResponse

    from starlite.types import ASGIApp, ExceptionHandlersMap, Receive, Scope, Send
    from starlite.types.asgi_types import WebSocketCloseEvent


class ExceptionHandlerMiddleware:
    def __init__(self, app: "ASGIApp", debug: bool, exception_handlers: "ExceptionHandlersMap") -> None:
        """This middleware is used to wrap an ASGIApp inside a try catch block
        and handles any exceptions raised.

        Notes:
            * It's used in multiple layers of Starlite.

        Args:
            app: The 'next' ASGI app to call.
            debug: Whether 'debug' mode is enabled
            exception_handlers: A dictionary mapping status codes and/or exception types to handler functions.
        """
        self.app = app
        self.exception_handlers = exception_handlers
        self.debug = debug

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        try:
            await self.app(scope, receive, send)
        except Exception as e:  # pylint: disable=broad-except
            starlite_app = scope["app"]
            for hook in starlite_app.after_exception:
                await hook(e, scope, starlite_app.state)

            if scope["type"] == ScopeType.HTTP:
                exception_handler = (
                    get_exception_handler(self.exception_handlers, e) or self.default_http_exception_handler
                )
                response = exception_handler(Request(scope=scope, receive=receive, send=send), e)
                await response(scope=scope, receive=receive, send=send)  # type: ignore[arg-type]
                return

            if isinstance(e, WebSocketException):
                code = e.code
                reason = e.detail
            elif isinstance(e, StarletteHTTPException):
                code = e.status_code + 4000
                reason = e.detail
            else:
                code = HTTP_500_INTERNAL_SERVER_ERROR + 4000
                reason = repr(e)
            event: "WebSocketCloseEvent" = {"type": "websocket.close", "code": code, "reason": reason}
            await send(event)

    def default_http_exception_handler(self, request: Request, exc: Exception) -> "StarletteResponse":
        """Default handler for exceptions subclassed from HTTPException."""
        status_code = exc.status_code if isinstance(exc, StarletteHTTPException) else HTTP_500_INTERNAL_SERVER_ERROR
        if status_code == HTTP_500_INTERNAL_SERVER_ERROR and self.debug:
            # in debug mode, we just use the serve_middleware to create an HTML formatted response for us
            server_middleware = ServerErrorMiddleware(app=self)  # type: ignore[arg-type]
            return server_middleware.debug_response(request=request, exc=exc)  # type: ignore[arg-type]
        return create_exception_response(exc)



from copy import copy
from typing import Any, Dict, Iterator, MutableMapping, Optional


class State(MutableMapping[str, Any]):

    __slots__ = ("_state",)

    _state: Dict[str, Any]

    def __init__(self, state: Optional[Dict[str, Any]] = None) -> None:
        """An object meant to store arbitrary state. It can be accessed using
        dot notation while exposing dict like functionalities.

        Args:
             state: An optional string keyed dict for the initial state.

        Examples:

        ```python
        from starlite import State

        state_dict = {"first": 1, "second": 2, "third": 3, "fourth": 4}
        state = State(state_dict)

        # state can be accessed using '.' notation
        assert state.fourth == 4
        del state.fourth

        # state implements the Mapping type:
        assert len(state) == 3
        assert "first" in state
        assert not "fourth" in state
        assert state["first"] == 1
        assert [(k, v) for k, v in state.items()] == [("first", 1), ("second", 2), ("third", 3)]

        state["fourth"] = 4
        assert "fourth" in state
        del state["fourth"]

        # state implements __bool__
        assert state  # state is true when it has values.
        assert not State()  # state is empty when it has no values.

        # it has shallow copy
        copied_state = state.copy()
        del copied_state.first
        assert state.first

        # it has a 'dict' method to retrieve a shallow copy of the underlying dict
        inner_dict = state.dict()
        assert inner_dict == state_dict
        ```
        """
        super().__setattr__("_state", state if state is not None else {})

    def __bool__(self) -> bool:
        """

        Returns:
            Boolean indicating whether the wrapped dict instance has values.
        """
        return bool(self._state)

    def __getitem__(self, key: str) -> Any:
        """

        Args:
            key: Key to access.

        Raises:
            KeyError

        Returns:
            A value from the wrapped state instance.
        """
        return self._state[key]

    def __delitem__(self, key: str) -> None:
        """
        Args:
            key: Key to access.

        Raises:
            KeyError

        Returns:
            None
        """
        del self._state[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """

        Args:
            key: Key to set.
            value: Value to set.

        Returns:
            None
        """
        self._state[key] = value

    def __iter__(self) -> Iterator[str]:
        """

        Returns:
            An iterator iterating the wrapped state dict.
        """
        return iter(self._state)

    def __len__(self) -> int:
        """

        Returns:
            The length of the wrapped state dict.
        """
        return len(self._state)

    def __setattr__(self, key: str, value: Any) -> None:
        """

        Args:
            key: Key to set.
            value: Value to set.

        Returns:
            None
        """
        self._state[key] = value

    def __getattr__(self, key: str) -> Any:
        """

        Args:
            key: Key to retrieve

        Raises:
            AttributeError: if the given attribute is not set.

        Returns:
            Retrieves the value for the corresponding key from the wrapped state object.
        """
        try:
            return self._state[key]
        except KeyError as e:
            raise AttributeError(f"State instance has no attribute '{key}'") from e

    def __delattr__(self, key: str) -> None:
        """

        Args:
            key: Key to delete

        Raises:
            AttributeError: if the given attribute is not set.

        Returns:
            Deletes the value from the key from the wrapped state object.
        """
        try:
            del self._state[key]
        except KeyError as e:
            raise AttributeError(f"State instance has no attribute '{key}'") from e

    def __copy__(self) -> "State":
        """Returns a shallow copy of the given state object.

        Customizes how the builtin "copy" function will work.
        """
        return self.__class__(copy(self._state))

    def copy(self) -> "State":
        """

        Returns:
            A shallow copy of 'self'.
        """
        return copy(self)

    def dict(self) -> Dict[str, Any]:
        """

        Returns:
            A shallow copy of the wrapped dict.
        """
        return copy(self._state)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode

from orjson import dumps, loads
from pydantic import BaseModel

from starlite.app import Starlite
from starlite.connection import Request
from starlite.enums import HttpMethod, ParamType, RequestEncodingType, ScopeType
from starlite.exceptions import MissingDependencyException
from starlite.handlers.http import get
from starlite.types import HTTPScope, RouteHandlerType
from starlite.types.asgi_types import ASGIVersion
from starlite.utils import default_serializer

if TYPE_CHECKING:
    from starlite.datastructures.cookie import Cookie

try:
    from httpx._content import (
        encode_json,
        encode_multipart_data,
        encode_urlencoded_data,
    )
    from httpx._types import FileTypes  # noqa: TC002
except ImportError as e:
    raise MissingDependencyException(
        "To use starlite.testing, install starlite with 'testing' extra, e.g. `pip install starlite[testing]`"
    ) from e


@get("/")
def _default_route_handler() -> None:
    ...


class RequestFactory:
    def __init__(
        self,
        app: Starlite = Starlite(route_handlers=[_default_route_handler]),
        server: str = "test.org",
        port: int = 3000,
        root_path: str = "",
        scheme: str = "http",
    ) -> None:
        """A factory object to create [Request][starlite.connection.Request]
        instances.

        Args:
             app: An instance of [Starlite][starlite.app.Starlite] to set as `request.scope["app"]`.
             server: The server's domain.
             port: The server's port.
             root_path: Root path for the server.
             scheme: Scheme for the server.

        Examples:

        ```python
        from starlite import RequestEncodingType, Starlite
        from starlite.testing import RequestFactory

        from tests import PersonFactory

        my_app = Starlite(route_handlers=[])
        my_server = "starlite.org"

        # Create a GET request
        query_params = {"id": 1}
        get_user_request = RequestFactory(app=my_app, server=my_server).get(
            "/person", query_params=query_params
        )

        # Create a POST request
        new_person = PersonFactory.build()
        create_user_request = RequestFactory(app=my_app, server=my_server).post(
            "/person", data=person
        )

        # Create a request with a special header
        headers = {"header1": "value1"}
        request_with_header = RequestFactory(app=my_app, server=my_server).get(
            "/person", query_params=query_params, headers=headers
        )

        # Create a request with a media type
        request_with_media_type = RequestFactory(app=my_app, server=my_server).post(
            "/person", data=person, request_media_type=RequestEncodingType.MULTI_PART
        )
        ```
        """

        self.app = app
        self.server = server
        self.port = port
        self.root_path = root_path
        self.scheme = scheme

    def _create_scope(
        self,
        path: str,
        http_method: HttpMethod,
        session: Optional[Dict[str, Any]] = None,
        user: Any = None,
        auth: Any = None,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
        state: Optional[Dict[str, Any]] = None,
        path_params: Optional[Dict[str, str]] = None,
        http_version: Optional[str] = "1.1",
        route_handler: Optional[RouteHandlerType] = None,
    ) -> HTTPScope:
        """Create the scope for the [Request][starlite.connection.Request].

        Args:
            path: The request's path.
            http_method: The request's HTTP method.
            session: A dictionary of session data.
            user: A value for `request.scope["user"]`.
            auth: A value for `request.scope["auth"]`.
            query_params: A dictionary of values from which the request's query will be generated.
            state: Arbitrary request state.
            path_params: A string keyed dictionary of path parameter values.
            http_version: HTTP version. Defaults to "1.1".
            route_handler: A route handler instance or method. If not provided a default handler is set.

        Returns:
            A dictionary that can be passed as a scope to the [Request][starlite.connection.Request] ctor.
        """
        if session is None:
            session = {}

        if state is None:
            state = {}

        if path_params is None:
            path_params = {}

        return HTTPScope(
            type=ScopeType.HTTP,
            method=http_method.value,
            scheme=self.scheme,
            server=(self.server, self.port),
            root_path=self.root_path.rstrip("/"),
            path=path,
            headers=[],
            app=self.app,
            session=session,
            user=user,
            auth=auth,
            query_string=urlencode(query_params, doseq=True).encode() if query_params else b"",
            path_params=path_params,
            client=(self.server, self.port),
            state=state,
            asgi=ASGIVersion(spec_version="3.0", version="3.0"),
            http_version=http_version or "1.1",
            raw_path=path.encode("ascii"),
            route_handler=route_handler or _default_route_handler,
            extensions={},
        )

    @classmethod
    def _create_cookie_header(
        cls, headers: Dict[str, str], cookies: Optional[Union[List["Cookie"], str]] = None
    ) -> None:
        """Create the cookie header and add it to the `headers` dictionary.

        Args:
            headers: A dictionary of headers, the cookie header will be added to it.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.
        """

        if not cookies:
            return

        if isinstance(cookies, list):
            cookie_header = "; ".join(cookie.to_header(header="") for cookie in cookies)
            headers[ParamType.COOKIE] = cookie_header
        elif isinstance(cookies, str):
            headers[ParamType.COOKIE] = cookies

    def _build_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Union[List["Cookie"], str]] = None,
    ) -> List[Tuple[bytes, bytes]]:
        """Build a list of encoded headers that can be passed to the request
        scope.

        Args:
            headers: A dictionary of headers.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.

        Returns:
            A list of encoded headers that can be passed to the request scope.
        """

        headers = headers or {}
        self._create_cookie_header(headers, cookies)
        return [
            ((key.lower()).encode("latin-1", errors="ignore"), value.encode("latin-1", errors="ignore"))
            for key, value in headers.items()
        ]

    def _create_request_with_data(
        self,
        http_method: HttpMethod,
        path: str,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Union[List["Cookie"], str]] = None,
        session: Optional[Dict[str, Any]] = None,
        user: Any = None,
        auth: Any = None,
        request_media_type: RequestEncodingType = RequestEncodingType.JSON,
        data: Optional[Union[Dict[str, Any], "BaseModel"]] = None,
        files: Optional[Union[Dict[str, FileTypes], List[Tuple[str, FileTypes]]]] = None,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
        state: Optional[Dict[str, Any]] = None,
        path_params: Optional[Dict[str, str]] = None,
        http_version: Optional[str] = "1.1",
        route_handler: Optional[RouteHandlerType] = None,
    ) -> Request[Any, Any]:
        """Create a [Request][starlite.connection.Request] instance that has
        body (data)

        Args:
            http_method: The request's HTTP method.
            path: The request's path.
            headers: A dictionary of headers.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.
            session: A dictionary of session data.
            user: A value for `request.scope["user"]`
            auth: A value for `request.scope["auth"]`
            request_media_type: The 'Content-Type' header of the request.
            data: A value for the request's body. Can be either a pydantic model instance
                or a string keyed dictionary.
            query_params: A dictionary of values from which the request's query will be generated.
            state: Arbitrary request state.
            path_params: A string keyed dictionary of path parameter values.
            http_version: HTTP version. Defaults to "1.1".
            route_handler: A route handler instance or method. If not provided a default handler is set.

        Returns:
            A [Request][starlite.connection.Request] instance
        """

        scope = self._create_scope(
            path=path,
            http_method=http_method,
            session=session,
            user=user,
            auth=auth,
            query_params=query_params,
            state=state,
            path_params=path_params,
            http_version=http_version,
            route_handler=route_handler,
        )

        headers = headers or {}
        if data:
            if isinstance(data, BaseModel):
                data = data.dict()
            if request_media_type == RequestEncodingType.JSON:
                encoding_headers, stream = encode_json(data)
            elif request_media_type == RequestEncodingType.MULTI_PART:
                encoding_headers, stream = encode_multipart_data(data, files=files or [])
            else:
                encoding_headers, stream = encode_urlencoded_data(loads(dumps(data, default=default_serializer)))
            headers.update(encoding_headers)
            body = b""
            for chunk in stream:
                body += chunk
            scope["_body"] = body  # type: ignore[typeddict-item]
        self._create_cookie_header(headers, cookies)
        scope["headers"] = self._build_headers(headers)
        return Request(scope=scope)

    def get(
        self,
        path: str = "/",
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Union[List["Cookie"], str]] = None,
        session: Optional[Dict[str, Any]] = None,
        user: Any = None,
        auth: Any = None,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
        state: Optional[Dict[str, Any]] = None,
        path_params: Optional[Dict[str, str]] = None,
        http_version: Optional[str] = "1.1",
        route_handler: Optional[RouteHandlerType] = None,
    ) -> Request[Any, Any]:
        """Create a GET [Request][starlite.connection.Request] instance.

        Args:
            path: The request's path.
            headers: A dictionary of headers.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.
            session: A dictionary of session data.
            user: A value for `request.scope["user"]`.
            auth: A value for `request.scope["auth"]`.
            query_params: A dictionary of values from which the request's query will be generated.
            state: Arbitrary request state.
            path_params: A string keyed dictionary of path parameter values.
            http_version: HTTP version. Defaults to "1.1".
            route_handler: A route handler instance or method. If not provided a default handler is set.

        Returns:
            A [Request][starlite.connection.Request] instance
        """

        scope = self._create_scope(
            path=path,
            http_method=HttpMethod.GET,
            session=session,
            user=user,
            auth=auth,
            query_params=query_params,
            state=state,
            path_params=path_params,
            http_version=http_version,
            route_handler=route_handler,
        )

        scope["headers"] = self._build_headers(headers, cookies)
        return Request(scope=scope)

    def post(
        self,
        path: str = "/",
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Union[List["Cookie"], str]] = None,
        session: Optional[Dict[str, Any]] = None,
        user: Any = None,
        auth: Any = None,
        request_media_type: RequestEncodingType = RequestEncodingType.JSON,
        data: Optional[Union[Dict[str, Any], "BaseModel"]] = None,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
        state: Optional[Dict[str, Any]] = None,
        path_params: Optional[Dict[str, str]] = None,
        http_version: Optional[str] = "1.1",
        route_handler: Optional[RouteHandlerType] = None,
    ) -> Request[Any, Any]:
        """Create a POST [Request][starlite.connection.Request] instance.

        Args:
            path: The request's path.
            headers: A dictionary of headers.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.
            session: A dictionary of session data.
            user: A value for `request.scope["user"]`.
            auth: A value for `request.scope["auth"]`.
            request_media_type: The 'Content-Type' header of the request.
            data: A value for the request's body. Can be either a pydantic model instance
                or a string keyed dictionary.
            query_params: A dictionary of values from which the request's query will be generated.
            state: Arbitrary request state.
            path_params: A string keyed dictionary of path parameter values.
            http_version: HTTP version. Defaults to "1.1".
            route_handler: A route handler instance or method. If not provided a default handler is set.

        Returns:
            A [Request][starlite.connection.Request] instance
        """

        return self._create_request_with_data(
            auth=auth,
            cookies=cookies,
            data=data,
            headers=headers,
            http_method=HttpMethod.POST,
            path=path,
            query_params=query_params,
            request_media_type=request_media_type,
            session=session,
            user=user,
            state=state,
            path_params=path_params,
            http_version=http_version,
            route_handler=route_handler,
        )

    def put(
        self,
        path: str = "/",
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Union[List["Cookie"], str]] = None,
        session: Optional[Dict[str, Any]] = None,
        user: Any = None,
        auth: Any = None,
        request_media_type: RequestEncodingType = RequestEncodingType.JSON,
        data: Optional[Union[Dict[str, Any], "BaseModel"]] = None,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
        state: Optional[Dict[str, Any]] = None,
        path_params: Optional[Dict[str, str]] = None,
        http_version: Optional[str] = "1.1",
        route_handler: Optional[RouteHandlerType] = None,
    ) -> Request[Any, Any]:
        """Create a PUT [Request][starlite.connection.Request] instance.

        Args:
            path: The request's path.
            headers: A dictionary of headers.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.
            session: A dictionary of session data.
            user: A value for `request.scope["user"]`.
            auth: A value for `request.scope["auth"]`.
            request_media_type: The 'Content-Type' header of the request.
            data: A value for the request's body. Can be either a pydantic model instance
                or a string keyed dictionary.
            query_params: A dictionary of values from which the request's query will be generated.
            state: Arbitrary request state.
            path_params: A string keyed dictionary of path parameter values.
            http_version: HTTP version. Defaults to "1.1".
            route_handler: A route handler instance or method. If not provided a default handler is set.

        Returns:
            A [Request][starlite.connection.Request] instance
        """

        return self._create_request_with_data(
            auth=auth,
            cookies=cookies,
            data=data,
            headers=headers,
            http_method=HttpMethod.PUT,
            path=path,
            query_params=query_params,
            request_media_type=request_media_type,
            session=session,
            user=user,
            state=state,
            path_params=path_params,
            http_version=http_version,
            route_handler=route_handler,
        )

    def patch(
        self,
        path: str = "/",
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Union[List["Cookie"], str]] = None,
        session: Optional[Dict[str, Any]] = None,
        user: Any = None,
        auth: Any = None,
        request_media_type: RequestEncodingType = RequestEncodingType.JSON,
        data: Optional[Union[Dict[str, Any], "BaseModel"]] = None,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
        state: Optional[Dict[str, Any]] = None,
        path_params: Optional[Dict[str, str]] = None,
        http_version: Optional[str] = "1.1",
        route_handler: Optional[RouteHandlerType] = None,
    ) -> Request[Any, Any]:
        """Create a PATCH [Request][starlite.connection.Request] instance.

        Args:
            path: The request's path.
            headers: A dictionary of headers.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.
            session: A dictionary of session data.
            user: A value for `request.scope["user"]`.
            auth: A value for `request.scope["auth"]`.
            request_media_type: The 'Content-Type' header of the request.
            data: A value for the request's body. Can be either a pydantic model instance
                or a string keyed dictionary.
            query_params: A dictionary of values from which the request's query will be generated.
            state: Arbitrary request state.
            path_params: A string keyed dictionary of path parameter values.
            http_version: HTTP version. Defaults to "1.1".
            route_handler: A route handler instance or method. If not provided a default handler is set.

        Returns:
            A [Request][starlite.connection.Request] instance
        """

        return self._create_request_with_data(
            auth=auth,
            cookies=cookies,
            data=data,
            headers=headers,
            http_method=HttpMethod.PATCH,
            path=path,
            query_params=query_params,
            request_media_type=request_media_type,
            session=session,
            user=user,
            state=state,
            path_params=path_params,
            http_version=http_version,
            route_handler=route_handler,
        )

    def delete(
        self,
        path: str = "/",
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Union[List["Cookie"], str]] = None,
        session: Optional[Dict[str, Any]] = None,
        user: Any = None,
        auth: Any = None,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
        state: Optional[Dict[str, Any]] = None,
        path_params: Optional[Dict[str, str]] = None,
        http_version: Optional[str] = "1.1",
        route_handler: Optional[RouteHandlerType] = None,
    ) -> Request[Any, Any]:
        """Create a POST [Request][starlite.connection.Request] instance.

        Args:
            path: The request's path.
            headers: A dictionary of headers.
            cookies: A string representing the cookie header or a list of "Cookie" instances.
                This value can include multiple cookies.
            session: A dictionary of session data.
            user: A value for `request.scope["user"]`.
            auth: A value for `request.scope["auth"]`.
            query_params: A dictionary of values from which the request's query will be generated.
            state: Arbitrary request state.
            path_params: A string keyed dictionary of path parameter values.
            http_version: HTTP version. Defaults to "1.1".
            route_handler: A route handler instance or method. If not provided a default handler is set.

        Returns:
            A [Request][starlite.connection.Request] instance
        """

        scope = self._create_scope(
            path=path,
            http_method=HttpMethod.DELETE,
            session=session,
            user=user,
            auth=auth,
            query_params=query_params,
            state=state,
            path_params=path_params,
            http_version=http_version,
            route_handler=route_handler,
        )
        scope["headers"] = self._build_headers(headers, cookies)
        return Request(scope=scope)



from typing import TYPE_CHECKING, Any, Dict, Optional

from starlite_multipart.datastructures import UploadFile as MultipartUploadFile

from starlite.openapi.enums import OpenAPIType

if TYPE_CHECKING:
    from pydantic.fields import ModelField


class UploadFile(MultipartUploadFile):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any], field: Optional["ModelField"]) -> None:
        """Creates a pydantic JSON schema.

        Args:
            field_schema: The schema being generated for the field.
            field: the model class field.

        Returns:
            None
        """
        if field:
            field_schema.update({"type": OpenAPIType.STRING.value, "contentMediaType": "application/octet-stream"})

    def __repr__(self) -> str:
        return f"{self.filename} - {self.content_type}"



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from starlite.types import Scope, Serializer


def get_serializer_from_scope(scope: "Scope") -> Optional["Serializer"]:
    """
    Utility that returns a serializer given a scope object.
    Args:
        scope: The ASGI connection scope.

    Returns:
        A serializer function
    """

    route_handler = scope["route_handler"]
    if hasattr(route_handler, "resolve_response_class"):
        return route_handler.resolve_response_class().serializer  # type: ignore

    app = scope["app"]
    return app.response_class.serializer if app.response_class else None



import re
from inspect import isawaitable
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Type, Union

from orjson import OPT_OMIT_MICROSECONDS, OPT_SERIALIZE_NUMPY, dumps
from pydantic import BaseModel

from starlite.enums import ScopeType
from starlite.middleware.base import DefineMiddleware, MiddlewareProtocol
from starlite.utils import default_serializer, get_serializer_from_scope
from starlite.utils.extractors import (
    ConnectionDataExtractor,
    RequestExtractorField,
    ResponseDataExtractor,
    ResponseExtractorField,
)

if TYPE_CHECKING:
    from starlite.connection import Request
    from starlite.types import ASGIApp, Logger, Message, Receive, Scope, Send

try:
    from structlog.types import BindableLogger

    structlog_installed = True  # pylint: disable=invalid-name
except ImportError:
    BindableLogger = object  # type: ignore
    structlog_installed = False  # pylint: disable=invalid-name


class LoggingMiddleware(MiddlewareProtocol):
    __slots__ = ("config", "logger", "exclude", "request_extractor", "response_extractor", "is_struct_logger")

    logger: "Logger"

    def __init__(self, app: "ASGIApp", config: "LoggingMiddlewareConfig") -> None:
        """LoggingMiddleware.

        Args:
            app: The 'next' ASGI app to call.
            config: An instance of LoggingMiddlewareConfig.
        """
        self.app = app
        self.is_struct_logger = structlog_installed
        self.config = config
        self.exclude = (
            (re.compile("|".join(config.exclude)) if isinstance(config.exclude, list) else re.compile(config.exclude))
            if config.exclude
            else None
        )
        self.request_extractor = ConnectionDataExtractor(
            extract_body="body" in self.config.request_log_fields,
            extract_client="client" in self.config.request_log_fields,
            extract_content_type="content_type" in self.config.request_log_fields,
            extract_cookies="cookies" in self.config.request_log_fields,
            extract_headers="headers" in self.config.request_log_fields,
            extract_method="method" in self.config.request_log_fields,
            extract_path="path" in self.config.request_log_fields,
            extract_path_params="path_params" in self.config.request_log_fields,
            extract_query="query" in self.config.request_log_fields,
            extract_scheme="scheme" in self.config.request_log_fields,
            obfuscate_cookies=self.config.request_cookies_to_obfuscate,
            obfuscate_headers=self.config.request_headers_to_obfuscate,
            parse_body=self.is_struct_logger,
            parse_query=self.is_struct_logger,
        )
        self.response_extractor = ResponseDataExtractor(
            extract_body="body" in self.config.response_log_fields,
            extract_headers="headers" in self.config.response_log_fields,
            extract_status_code="status_code" in self.config.response_log_fields,
            obfuscate_cookies=self.config.response_cookies_to_obfuscate,
            obfuscate_headers=self.config.response_headers_to_obfuscate,
        )

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """

        if scope["type"] == ScopeType.HTTP and not (self.exclude and self.exclude.findall(scope["path"])):
            if not hasattr(self, "logger"):
                self.logger = scope["app"].get_logger(self.config.logger_name)
                self.is_struct_logger = structlog_installed and isinstance(self.logger, BindableLogger)
            if self.config.request_log_fields:
                await self.log_request(scope=scope)
            if self.config.response_log_fields:
                send = self.create_send_wrapper(scope=scope, send=send)
        await self.app(scope, receive, send)

    async def log_request(self, scope: "Scope") -> None:
        """
        Handles extracting the request data and logging the message.
        Args:
            scope: The ASGI connection scope.

        Returns:
            None

        """
        extracted_data = await self.extract_request_data(request=scope["app"].request_class(scope))
        self.log_message(values=extracted_data)

    def log_response(self, scope: "Scope") -> None:
        """Handles extracting the response data and logging the message.

        Args:
            scope: The ASGI connection scope.

        Returns:
            None
        """
        extracted_data = self.extract_response_data(scope=scope)
        self.log_message(values=extracted_data)

    def log_message(self, values: Dict[str, Any]) -> None:
        """

        Args:
            values: Extract values to log.

        Returns:
            None

        """
        message = values.pop("message")
        if self.is_struct_logger:

            self.logger.info(message, **values)
        else:
            self.logger.info(f"{message}: " + ", ".join([f"{key}={value}" for key, value in values.items()]))

    async def extract_request_data(self, request: "Request") -> Dict[str, Any]:
        """Creates a dictionary of values for the message.

        Args:
            request: A [Request][starlite.connection.Request] instance.
        Returns:
            An OrderedDict.
        """

        data: Dict[str, Any] = {"message": self.config.request_log_message}
        serializer = get_serializer_from_scope(request.scope) or default_serializer
        extracted_data = self.request_extractor(connection=request)
        for key in self.config.request_log_fields:
            value = extracted_data.get(key)
            if isawaitable(value):
                value = await value
            if not self.is_struct_logger and isinstance(value, (dict, list, tuple, set)):
                value = dumps(value, default=serializer, option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS)
            if isinstance(value, bytes):
                value = value.decode("utf-8")
            data[key] = value
        return data

    def extract_response_data(self, scope: "Scope") -> Dict[str, Any]:
        """
        Extracts data from the response.
        Args:
            scope: The ASGI connection scope.

        Returns:
            An OrderedDict.
        """
        data: Dict[str, Any] = {"message": self.config.response_log_message}
        serializer = get_serializer_from_scope(scope) or default_serializer
        extracted_data = self.response_extractor(
            messages=(
                scope["state"]["http.response.start"],
                scope["state"]["http.response.body"],
            )
        )
        for key in self.config.response_log_fields:
            value = extracted_data.get(key)
            if not self.is_struct_logger and isinstance(value, (dict, list, tuple, set)):
                value = dumps(value, default=serializer, option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS)
            if isinstance(value, bytes):
                value = value.decode("utf-8")
            data[key] = value
        return data

    def create_send_wrapper(self, scope: "Scope", send: "Send") -> "Send":
        """Creates a 'send' wrapper, which handles logging response data.

        Args:
            scope: The ASGI connection scope.
            send: The ASGI send function.

        Returns:
            An ASGI send function.
        """

        async def send_wrapper(message: "Message") -> None:
            if message["type"] == "http.response.start":
                scope["state"]["http.response.start"] = message
            elif message["type"] == "http.response.body":
                scope["state"]["http.response.body"] = message
                self.log_response(scope=scope)
            await send(message)

        return send_wrapper


class LoggingMiddlewareConfig(BaseModel):
    exclude: Optional[Union[str, List[str]]] = None
    """
    List of paths to exclude from logging.
    """
    logger_name: str = "starlite"
    """
    Name of the logger to retrieve using `app.get_logger("<name>")`.
    """
    request_cookies_to_obfuscate: Set[str] = {"session"}
    """
    Request cookie keys to obfuscate. Obfuscated values are replaced with '*****'.
    """
    request_headers_to_obfuscate: Set[str] = {"Authorization", "X-API-KEY"}
    """
    Request header keys to obfuscate. Obfuscated values are replaced with '*****'.
    """
    response_cookies_to_obfuscate: Set[str] = {"session"}
    """
    Response cookie keys to obfuscate. Obfuscated values are replaced with '*****'.
    """
    response_headers_to_obfuscate: Set[str] = {"Authorization", "X-API-KEY"}
    """
    Response header keys to obfuscate. Obfuscated values are replaced with '*****'.
    """
    request_log_message: str = "HTTP Request"
    """
    Log message to prepend when logging a request.
    """
    response_log_message: str = "HTTP Response"
    """
    Log message to prepend when logging a response.
    """
    request_log_fields: Iterable[RequestExtractorField] = (
        "path",
        "method",
        "content_type",
        "headers",
        "cookies",
        "query",
        "path_params",
        "body",
    )
    """
    Fields to extract and log from the request.

    Notes:
        -  The order of fields in the iterable determines the order of the log message logged out.
            Thus, re-arranging the log-message is as simple as changing the iterable.
        -  To turn off logging of requests, use and empty iterable.

    """
    response_log_fields: Iterable[ResponseExtractorField] = (
        "status_code",
        "cookies",
        "headers",
        "body",
    )
    """
    Fields to extract and log from the response. The order of fields in the iterable determines the order of the log message logged out.

    Notes:
        -  The order of fields in the iterable determines the order of the log message logged out.
            Thus, re-arranging the log-message is as simple as changing the iterable.
        -  To turn off logging of responses, use and empty iterable.
    """
    middleware_class: Type[LoggingMiddleware] = LoggingMiddleware
    """
    Middleware class to use. Should be a subclass of [starlite.middleware.LoggingMiddleware].
    """

    @property
    def middleware(self) -> DefineMiddleware:
        """Use this property to insert the config into a middleware list on one
        of the application layers.

        Examples:

            ```python
            from starlite import Starlite, Request, LoggingConfig, get
            from starlite.middleware.logging import LoggingMiddlewareConfig

            logging_config = LoggingConfig()

            logging_middleware_config = LoggingMiddlewareConfig()


            @get("/")
            def my_handler(request: Request) -> None:
                ...


            app = Starlite(
                route_handlers=[my_handler],
                logging_config=logging_config,
                middleware=[logging_middleware_config.middleware],
            )
            ```

        Returns:
            An instance of DefineMiddleware including 'self' as the config kwarg value.
        """
        return DefineMiddleware(self.middleware_class, config=self)



from inspect import getmro
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union, cast

from pydantic import BaseModel
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from starlite.enums import MediaType
from starlite.exceptions.http_exceptions import HTTPException
from starlite.response import Response

if TYPE_CHECKING:
    from starlite.types import ExceptionHandler


def get_exception_handler(
    exception_handlers: Dict[Union[int, Type[Exception]], "ExceptionHandler"], exc: Exception
) -> Optional["ExceptionHandler"]:
    """Given a dictionary that maps exceptions and status codes to handler
    functions, and an exception, returns the appropriate handler if existing.

    Status codes are given preference over exception type.

    If no status code match exists, each class in the MRO of the exception type is checked and
    the first matching handler is returned.

    Finally, if a `500` handler is registered, it will be returned for any exception that isn't a
    subclass of `HTTPException`.

    Args:
        exception_handlers: Mapping of status codes and exception types to handlers.
        exc: Exception Instance to be resolved to a handler.

    Returns:
        Optional exception handler callable.
    """
    if not exception_handlers:
        return None
    if isinstance(exc, (StarletteHTTPException, HTTPException)) and exc.status_code in exception_handlers:
        return exception_handlers[exc.status_code]
    for cls in getmro(type(exc)):
        if cls in exception_handlers:
            return exception_handlers[cast("Type[Exception]", cls)]
    if (
        not isinstance(exc, (StarletteHTTPException, HTTPException))
        and HTTP_500_INTERNAL_SERVER_ERROR in exception_handlers
    ):
        return exception_handlers[HTTP_500_INTERNAL_SERVER_ERROR]
    return None


class ExceptionResponseContent(BaseModel):
    status_code: int
    """Exception status code."""
    detail: str
    """Exception details or message."""
    headers: Optional[Dict[str, str]] = None
    """Headers to attach to the response."""
    extra: Optional[Union[Dict[str, Any], List[Any]]] = None
    """An extra mapping to attach to the exception."""


def create_exception_response(exc: Exception) -> Response:
    """Constructs a response from an exception.

    For instances of either `starlite.exceptions.HTTPException` or `starlette.exceptions.HTTPException` the response
    status code is drawn from the exception, otherwise response status is `HTTP_500_INTERNAL_SERVER_ERROR`.

    Args:
        exc: An exception.

    Returns:
        Response: HTTP response constructed from exception details.
    """
    if isinstance(exc, (HTTPException, StarletteHTTPException)):
        content = ExceptionResponseContent(detail=exc.detail, status_code=exc.status_code)
        if isinstance(exc, HTTPException):
            content.extra = exc.extra
    else:
        content = ExceptionResponseContent(detail=repr(exc), status_code=HTTP_500_INTERNAL_SERVER_ERROR)
    return Response(
        media_type=MediaType.JSON,
        content=content.dict(exclude_none=True),
        status_code=content.status_code,
        headers=exc.headers if isinstance(exc, (HTTPException, StarletteHTTPException)) else None,
    )



from http.cookies import SimpleCookie
from typing import Any, Optional

from pydantic import BaseModel
from typing_extensions import Literal


class Cookie(BaseModel):
    """Container class for defining a cookie using the 'Set-Cookie' header.

    See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie for more details regarding this header.
    """

    key: str
    """Key for the cookie."""
    value: Optional[str] = None
    """Value for the cookie, if none given defaults to empty string."""
    max_age: Optional[int] = None
    """Maximal age of the cookie before its invalidated."""
    expires: Optional[int] = None
    """Expiration date as unix MS timestamp."""
    path: str = "/"
    """Path fragment that must exist in the request url for the cookie to be valid. Defaults to '/'."""
    domain: Optional[str] = None
    """Domain for which the cookie is valid."""
    secure: Optional[bool] = None
    """Https is required for the cookie."""
    httponly: Optional[bool] = None
    """Forbids javascript to access the cookie via 'Document.cookie'."""
    samesite: Literal["lax", "strict", "none"] = "lax"
    """Controls whether or not a cookie is sent with cross-site requests. Defaults to 'lax'."""
    description: Optional[str] = None
    """Description of the response cookie header for OpenAPI documentation"""
    documentation_only: bool = False
    """Defines the Cookie instance as for OpenAPI documentation purpose only"""

    def to_header(self, **kwargs: Any) -> str:
        """Return a string representation suitable to be sent as HTTP headers.

        Args:
            **kwargs: Passed to [SimpleCookie][http.cookies.SimpleCookie]
        """

        simple_cookie: SimpleCookie = SimpleCookie()
        simple_cookie[self.key] = self.value or ""
        if self.max_age:
            simple_cookie[self.key]["max-age"] = self.max_age
        cookie_dict = self.dict()
        for key in ("expires", "path", "domain", "secure", "httponly", "samesite"):
            value = cookie_dict[key]
            if value is not None:
                simple_cookie[self.key][key] = value
        return simple_cookie.output(**kwargs).strip()
